{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE OverloadedStrings #-}
module Language.Java.Classfile.ConstantPool (ConstantPool(..)) where
import Data.Word (Word16)
import Data.Array.IArray (Array, listArray)
import Language.Java.Classfile.ConstantPool.Entry (Entry, StorageCount (..), storageCount)
import Language.Java.Classfile.Extractable (Extractable (extract))
import Language.Java.Classfile.Extract (Extract, traceIndex, traceType)
import Control.Monad (forM)
import qualified Data.Text as Text

newtype ConstantPool = ConstantPool (Array Word16 Entry)
  deriving stock (Show)

instance Extractable ConstantPool where
  extract :: Extract ConstantPool
  extract = do
    count <- extract @Word16

    let typeName = Text.concat
          [ "ConstantPool ("
          , Text.pack . show $ count
          , ")"
          ]

    let extractElements index end
          | index == end = pure []
          | otherwise = do
            element <- traceIndex index extract
            case storageCount element of
              Once  -> (element:) <$> extractElements (succ index) end
              Twice -> ([element, element] <>) <$> extractElements (index + 2) end

    traceType typeName $ do
      elements <- extractElements 1 (fromIntegral count)
      pure . ConstantPool $ listArray (1, count - 1) elements
